#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

typedef long long int64;

enum {
	MSD_MSG_STREAM_OPEN = 1,
	MSD_MSG_WRITE_DATA = 2,
	MSD_MSG_REQUEST_STATUS_NOTIFICATION = 3
};

void msd_debug(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf (stderr, "[msd] ");
	vfprintf (stderr, fmt, ap);
	fprintf (stderr, "\n");
    va_end(ap);
}

#define arts_debug msd_debug 
#define arts_warning msd_debug 

typedef struct _MsdConnection MsdConnection;

struct _MsdConnection {
	int fd;
	int nextStream;
	int nextMsgId;
};

typedef struct _MsdMessage MsdMessage;

struct _MsdMessage {
	int rpos;
	int wpos;
	int msgid;
	char buffer[8192];
};

static struct sockaddr_in *parse_tcp_url(const char *url)
{
	static struct sockaddr_in addr;
	char *work, *type, *host, *port;
	int portno;
    struct hostent *server;

	work = strdup(url);

	type = strtok(work,":");
	if(type == 0 || strcmp(type,"tcp") != 0) {
		free(work);
		return 0;
	}

	host = strtok(NULL,":");
	if(host == 0) {
		free(work);
		return 0;
	}

	port = strtok(NULL,":\n");
	if(port == 0) {
		free(work);
		return 0;
	}

	portno = atoi(port);
	if(portno < 1 || portno > 65535) {
		free(work);
		return 0;
	}

	server = gethostbyname(host);
   	if(server == 0)
	{
		arts_warning("parse_tcp_url: unknown host '%s'",host);
		free(work);
		return 0;
	}

	memset((void *) &addr, '\0', sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = *(u_long *)server->h_addr;
    addr.sin_port = htons(portno);

	free(work);
	return &addr;
}

MsdConnection *msd_connect(const char *url)
{
	struct sockaddr_in *remote_addr;
	struct linger lin;
	int my_socket;
	int on;
	int rc;
	MsdConnection *conn;

	if (!url)
		url = "tcp:localhost:17771";

	remote_addr = parse_tcp_url(url);
	if(remote_addr == 0)
	{
		arts_warning("tcp_connect: couldn't parse url %s",url);
		return 0;
	}

	my_socket = socket(AF_INET,SOCK_STREAM,0);
	if(my_socket < 0)
	{
		arts_warning("tcp_connect: unable to open socket for read");
		return 0;
	}

    lin.l_onoff=1;      /* block a closing socket for 1 second */
    lin.l_linger=100;   /* if data is waiting to be sent */
    if ( setsockopt( my_socket, SOL_SOCKET, SO_LINGER,
                     (char*)&lin, sizeof(struct linger) ) < 0 )
    {
        arts_warning("tcp_connect: unable to set socket linger value to %d",
                lin.l_linger);
		close(my_socket);
        return 0;
    }

#ifdef TCP_NODELAY
	// enable TCP sending without nagle algorithm - this sends out requests
	// faster, because  they are not queued in the hope that more data will
	// need to be sent soon
	on = 1;
	if(setsockopt(my_socket, IPPROTO_TCP, TCP_NODELAY, 
	              (char *)&on, sizeof(on)) < 0)
	{
		arts_debug("couldn't set TCP_NODELAY on socket %d\n", my_socket);
	}
#endif

	rc=connect(my_socket,(struct sockaddr *)remote_addr, sizeof(*remote_addr));
	if(rc != 0)
	{
		arts_warning("tcp_connect: can't connect to server (%s)", url);
		close(my_socket);
		return 0;
	}

	/* fixme: handle errors */
	conn = (MsdConnection *)malloc(sizeof(MsdConnection));
	conn->fd = my_socket;
	conn->nextStream = 1;
	conn->nextMsgId = 1;
	return conn;
}

int msd_message_get_int(MsdMessage *msg)
{
	int result = (msg->buffer[msg->rpos]   << 24)
			   + (msg->buffer[msg->rpos+1] << 16)
			   + (msg->buffer[msg->rpos+2] << 8)
	 		   +  msg->buffer[msg->rpos+3];
	msg->rpos += 4;
	return result;
}

int64 msd_message_get_int64(MsdMessage *msg)
{
	int64 result = ((int64)msg->buffer[msg->rpos]   << 56)
			     + ((int64)msg->buffer[msg->rpos+1] << 48)
			     + ((int64)msg->buffer[msg->rpos+2] << 40)
			     + ((int64)msg->buffer[msg->rpos+3] << 32)
			     + ((int64)msg->buffer[msg->rpos+4] << 24)
			     + ((int64)msg->buffer[msg->rpos+5] << 16)
			     + ((int64)msg->buffer[msg->rpos+6] << 8)
	 		     +  (int64)msg->buffer[msg->rpos+7];
	msg->rpos += 8;
	return result;
}

void msd_message_put_int(MsdMessage *msg, int i)
{
	msg->buffer[msg->wpos++] = ((i >> 24) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 16) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 8) & 0xff);
	msg->buffer[msg->wpos++] = (i & 0xff);
}

void msd_message_put_int64(MsdMessage *msg, int64 i)
{
	msg->buffer[msg->wpos++] = ((i >> 56) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 48) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 40) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 32) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 24) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 16) & 0xff);
	msg->buffer[msg->wpos++] = ((i >> 8) & 0xff);
	msg->buffer[msg->wpos++] = (i & 0xff);
}

void msd_message_put_bytes(MsdMessage *msg, const void *data, int len)
{
	msd_message_put_int(msg, len);
	memcpy(&msg->buffer[msg->wpos], data, len);
	msg->wpos += len;
}

void msd_message_init(MsdMessage *msg, int type)
{
	msg->rpos = 0;

	strcpy(msg->buffer,"MSD");
	msg->wpos = 4;
		
	msd_message_put_int(msg, type);
	msd_message_put_int(msg, 0); /* length */
	msd_message_put_int(msg, 0); /* msgid */
}

void msd_message_send(MsdConnection *conn, MsdMessage *msg)
{
	int length = msg->wpos;
	int rc;

	msg->wpos = 8;
	msg->msgid = conn->nextMsgId++;

	msd_message_put_int(msg, length);
	msd_message_put_int(msg, msg->msgid);

	arts_debug("sending message (length = %d)", length);
	rc = write(conn->fd, msg->buffer, length);
	assert(rc == length);
}

void msd_message_wait_reply(MsdConnection *conn, MsdMessage *msg)
{
	fd_set rfds;
	int retval;
	/*
	struct timeval tv;
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	*/

	for(;;)
	{
		FD_ZERO(&rfds);
		FD_SET(conn->fd, &rfds);

		printf("waiting for data\n");
		retval = select(conn->fd + 1, &rfds, NULL, NULL, NULL);
        if (retval == 1)
		{
			printf("got data\n");
			/* FIXME */
		}
        else
		{
			printf("no data\n");
		}
	}
}

int msd_open_stream(MsdConnection *conn, int rate, int channels, int format)
{
	MsdMessage msg;
	int stream = conn->nextStream++;
	int result;

	msd_message_init(&msg, MSD_MSG_STREAM_OPEN);
	msd_message_put_int(&msg, stream);
	msd_message_put_int(&msg, rate);
	msd_message_put_int(&msg, channels);
	msd_message_put_int(&msg, format);
	msd_message_send(conn, &msg);
	msd_message_wait_reply(conn, &msg);
	result = msd_message_get_int(&msg);

	return stream;
}

void msd_close_stream(MsdConnection *conn, int stream)
{
}

void msd_write_data(MsdConnection *conn, int stream,
		            int64 startpos, const void *data, int len)
{
	MsdMessage msg;
	int result;

	msd_message_init(&msg, MSD_MSG_WRITE_DATA);
	msd_message_put_int(&msg, stream);
	msd_message_put_int64(&msg, startpos);
	msd_message_put_bytes(&msg, data, len);
	msd_message_send(conn, &msg);
	msd_message_wait_reply(conn, &msg);
	result = msd_message_get_int(&msg);
}

/* synchronous version: TODO: write asynchronous version */
int64 msd_request_status_notification(MsdConnection *conn, int stream, int64 pos)
{
	MsdMessage msg;
	int64 realpos = 0;
	
	msd_message_init(&msg, MSD_MSG_REQUEST_STATUS_NOTIFICATION);
	msd_message_put_int(&msg, stream);
	msd_message_put_int64(&msg, pos);
	msd_message_send(conn, &msg);
	msd_message_wait_reply(conn, &msg);
	realpos = msd_message_get_int64(&msg);

	return realpos;
}

//void msd_read <<int stream>> <<int64 startpos>> <<int len>>

static void exitUsage(const char *progname)
{
	fprintf(stderr,"usage: %s [ options ] [ <filename> ]\n",progname);
	fprintf(stderr,"-r <samplingrate>   set samplingrate to use\n");
	fprintf(stderr,"-b <bits>           set number of bits (8 or 16)\n");
	fprintf(stderr,"-c <channels>       set number of channels (1 or 2)\n");
	fprintf(stderr,"-h                  display this help and exit\n");
	exit(1);	
}

int main(int argc, char ** argv)
{
	int cfgSamplingRate = 44100;
	int cfgBits = 16;
	int cfgChannels = 2;
	FILE *infile = stdin;
	int pfd;
	int packetsize;
	char *buffer;
	int size = 1024;
	int optch;
	MsdConnection *conn;
	int stream;
	int64 pos;

	conn = msd_connect(0);
	if(!conn)
	{
		arts_debug("connection failed");
		return 1;
	}
	sleep(1);

	while((optch = getopt(argc,argv,"r:b:c:h")) > 0)
	{
		switch(optch)
		{
			case 'r': cfgSamplingRate = atoi(optarg);
				break;
			case 'b': cfgBits = atoi(optarg);
				break;
			case 'c': cfgChannels = atoi(optarg);
				break;
			case 'h':
			default: 
					exitUsage(argc?argv[0]:"artsccat");
				break;
		}
	}

	if (optind < argc)
    {
		if(argv[optind] != "-")
		{
			infile = fopen(argv[optind],"w");
			if(!infile)
			{
				fprintf( stderr, "Can't open file '%s'.\n", argv[optind] );
				exit(1);
			}
		}
	}

	pfd = fileno( infile );

	stream = msd_open_stream( conn, cfgSamplingRate, cfgChannels, cfgBits );
	packetsize = 4096;
	buffer = (char *)malloc(packetsize);
	pos = 0;

	do {
		size = read( pfd, buffer, packetsize );

		//FIXME: need returncode?
		msd_write_data( conn, stream, pos, buffer, size );
		pos += size / cfgChannels * 8 / cfgBits;

		/* 250 ms of buffering */
		msd_request_status_notification (conn, stream,
				                         pos - (cfgSamplingRate * 250 / 1000));
	} while( size > 0 );

	msd_close_stream( conn, stream );
	pclose( infile );
	free(buffer);

	return 0;
}
